

#define __CANC__
/**********************************************************************
 *  can.c
 *	Funktion:
 *	Treiberinterface fuer eine 82C200-basierte CAN-Interfacehardware.
 *	Der Treiber arbeitet im Pollingverfahren, und gestattet das Lesen,
 *	Schreiben und Konfigurieren des CAN-Interfaces.
 *
 *  Bereitgestellte #defines:
 *	    HARDWARETYP	        - Einstellen der verwendeten Hardware.
 *
 *  Bereitgestellte Prozeduren:
 *      CAN_Init()          - Modul initialisieren fuer Empfang aller Msgs
 *      CAN_registerMsg()   - Empfangsfilter fuer Msg erweitern.
 *      CAN_Read()          - Empfangspuffer lesen und Status zurueck,
 *				evtl. Controller reaktivieren.
 *      CAN_Write()         - Sendepuffer schreiben und Status zurueck
 *      CAN_Status()        - Statusregister des CAN-Controllers
 *      CAN_Close()         - Controller deaktivieren.
 *
 *      Autor   : J. Hoppe,      (c) 1999 PEAK-System Technik
 *      Sprache : Borland-C
 *
 *
 *      Revision-History:
 *      ----------------------------------------------------------------
 *      1.6.1996:       Portierung nach PC und 82C200.
 *      ----------------------------------------------------------------
 *
 *      Literatur:
 *      ----------------------------------------------------------------
 *      [AMC]   Application of the P8xC592 microcontroller with
 *              CAN-interface.
 *              P. Buehring, P. Hank, Philips Application Note, 7/92
 *      [PLC]   CAN Physical Layer Concepts for the P8xC592.
 *              H. Eisele, Philips Application Note, 7/92
 *      [TDS]   80C592 Target Device Specification 1991 Vers. 2.1.
 *              Philips, August 1991
 *
 *
 *      Copyright      (c) 1999 PEAK-System Technik GmbH, Darmstadt.
 */

#include "can.h"        /* eigene Typendeklarationen */


/************************************************************************/
/***    Konstanten                                                    ***/
/************************************************************************/

/* Auswahl der anzusteuernden CAN-Hardware ISA-Einsteckkarte oder eine
 * der Donglevarianten) */

#undef ISACAN		/* nur eines dieser Symbole #definen ! */
#define DONGLEMUX
#undef DUNGLEEPP

#ifdef ISACAN
/***** Ansteuerung der ISA-Karte: ISAREG.C einbinden! *****/
/* Prototypen der Treiberroutinen in ISAREG.C */
unsigned char isa_regread(int basisadr, unsigned char reg_adr) ;
void isa_regwrite(int basisadr, unsigned char reg_adr, unsigned char reg_val) ;
/* Ansteuerung definieren */
#define CANHWINIT(adr)
#define CANREAD(adr) isa_regread(PortBase, (adr))
#define CANWRITE(adr,val) isa_regwrite( PortBase, (adr), (val))

#endif

#ifdef DONGLEMUX
/***** Ansteuerung des Dongles im MUX-Mode: DONGLEREG.C einbinden! *****/
/* Prototypen der MUX-Treiberroutinen in DOBGLEREG.C */
void mux_reginit(int basisadr) ;
unsigned char mux_regread(int basisadr, unsigned char reg_adr) ;
void mux_regwrite(int basisadr, unsigned char reg_adr, unsigned char reg_val) ;
/* Ansteuerung definieren */
#define CANHWINIT(adr) mux_reginit(adr)
#define CANREAD(adr) mux_regread(PortBase, (adr))
#define CANWRITE(adr,val) mux_regwrite( PortBase, (adr), (val))

#endif

#ifdef DONGLEEPP
/***** Ansteuerung des Dongles im EPP-Mode: DONGLEREG.C einbinden! *****/
/* Prototypen der EPP-Treiberroutinen in DONGLEREG.C */
void epp_reginit(int basisadr) ;
unsigned char epp_regread(int basisadr, unsigned char reg_adr) ;
void epp_regwrite(int basisadr, unsigned char reg_adr, unsigned char reg_val) ;
/* Ansteuerung definieren */
#define CANHWINIT(adr) epp_reginit(adr)
#define CANREAD(adr) epp_regread(PortBase, (adr))
#define CANWRITE(adr,val) epp_regwrite( PortBase, (adr), (val))


#endif



/**** Registeradressen des CAN_Controllers 82C200 ***/
#define CONTROL         0
#define COMMAND         1
#define STATUS          2
#define INTERRUPT       3
#define ACC_CODE        4       /* Acceptance Code */
#define ACC_MASK        5       /* Acceptance Mask */
#define TIMING0         6       /* Bus Timing 0 */
#define TIMING1         7       /* Bus Timing 1 */
#define OUTPUTCTRL      8       /* Output Control */

#define XMTID           10      /* Transmit Buffer: Identifier       */
#define XMTLEN          11      /*     RTR Bit und Data length Code  */
#define XMTBUF          12      /*     BYTE1 des data field's        */

#define RCVID           20      /* Receive Buffer: Identifier        */
#define RCVLEN          21      /*     RTR Bit und Data length Code  */
#define RCVBUF          22      /*     BYTE1 des data field's        */



/************************************************************************/
/***    Lokale Variablen                                              ***/
/************************************************************************/

static unsigned int	PortBase ;    /* Basisadresse der Hardwarem aus CAN_Init() */
static int 		nomsgregistered ;  /* Flag fuer  'CAN_registerMsg()' */
static int		controlleractive = 0 ;


/************************************************************************/
/***    Lokale Prozeduren dieses Moduls                               ***/
/************************************************************************/


/*** CANBITSET(): die Bits in 'mask' im Controllerregister 'reg_adr' setzen ***/
static void CANBITSET(unsigned char reg_adr, unsigned char mask) {
	/* Besonderheit des COMMAND-Registers: Bits 4..0 sind write only */
	if ( reg_adr == COMMAND )
		CANWRITE(COMMAND, (CANREAD(COMMAND) & 0xE0) | mask) ;
	else
		CANWRITE(reg_adr, CANREAD(reg_adr) | mask) ;
}

/*** CANBITCLR(): die Bits in 'mask' im Controllerregister 'reg_adr' loeschen ***/
static void CANBITCLR(unsigned char reg_adr, unsigned char mask) {
	/* Besonderheit des COMMAND-Registers: Bits 4..0 sind write only */
	if ( reg_adr == COMMAND )
		CANWRITE(COMMAND, (CANREAD(COMMAND) & 0xE0) & ~ mask) ;
	else
		CANWRITE(reg_adr, CANREAD(reg_adr) & ~ mask) ;
}



/************************************************************************/
/***    Global verwendbare Prozeduren (Methoden) dieses Moduls        ***/
/************************************************************************/


/************************************************************************
 *  CAN_Read()
 *  Polling-Routine:
 *      -  empfangene Message in den Puffer 'PCANMsg->' uebertragen
 *      -  Ueberwachung des Bus-Status und ggf. Reset des CAN-Controllers.
 *
 *  Algorithmus angelehnt an [AMC], S. 18.
 *
 *	Parameter:
 *      PCANMsg: Zeiger auf Messagepuffer fuer die Rueckgabe der
 *		empfangenen Message.
 *
 *  Rueckgabe: Bitmaske mit Fehlern.
 *		0 :	Kein Fehler, neue Message steht in 'PCANMsg->'.
 *		CAN_ERR_RCVEMPTY: es wurde keine neue Message empfangen.
 *		CAN_ERR_OVERRUN : eine empfangene Message wurde nicht rechtzeitig
 *				ausgelesen.
 *		CAN_ERR_BUSERROR: Controller meldet viele Busstoerungen.
 *		CAN_ERR_BUSOFF: Controller war Bus-Off und wurde wieder
 *			aktiviert.
 */
int     CAN_Read(TCANMsg *PCANMsg)
{
        int     received ;
        int	result = 0 ;

        /************************ Datenempfang **************************/
        received = 0 ;
        if ( CANREAD(STATUS) & 0x01 )  {  /* Receive buffer status == full ? */
                /***  empfange jede Message, die dem Akzeptanzfilter genuegt */
                received = 1 ;
                PCANMsg->ID = ((unsigned short int) CANREAD(RCVID) << 3) | (CANREAD(RCVLEN) >> 5) ;
                PCANMsg->LEN = CANREAD(RCVLEN) & 0x0f ;
                if (CANREAD(RCVLEN) & 0x10 )
                    PCANMsg->RTR = 1 ; /* RTR-Bit gesetzt! */
                else {
                    /* den ganzen Empfangspuffer nach CAN_RWBUFF bewegen */
                    PCANMsg->RTR = 0 ;
                    PCANMsg->DATA[0] = CANREAD(RCVBUF) ;
                    PCANMsg->DATA[1] = CANREAD(RCVBUF+1) ;
                    PCANMsg->DATA[2] = CANREAD(RCVBUF+2) ;
                    PCANMsg->DATA[3] = CANREAD(RCVBUF+3) ;
                    PCANMsg->DATA[4] = CANREAD(RCVBUF+4) ;
                    PCANMsg->DATA[5] = CANREAD(RCVBUF+5) ;
                    PCANMsg->DATA[6] = CANREAD(RCVBUF+6) ;
                    PCANMsg->DATA[7] = CANREAD(RCVBUF+7) ;
                }
                /* release receive buffer */
                CANBITSET(COMMAND,0x04) ;
                /* Evtl. 'Overrun'-Bedingung loeschen */
                if ( CANREAD(STATUS) & 0x02 ) {
                	result |= CAN_ERR_OVERRUN ;
                        CANBITSET(COMMAND, 0x08 ) ;
		}
        }

        /********************* Fehlerverarbeitung ***********************/
        if ( CANREAD(STATUS) & 0x80 ) {
                /*** Bus-Off: Controller is NOT involved in bus activities */
		result |= CAN_ERR_BUSOFF ;
                /*** Reset request auf 'absent' */
                CANBITCLR(CONTROL, 0x01) ;
                /* Der Controller geht wieder 'On-Bus' nach dem Empfang
                 * von 128 * 11 aufeinanderfolgenden rezessiven Bits.
                 */
        } else {
                /*** Bus-On: Controller is involved in bus activities **/

                if ( CANREAD(STATUS) & 0x40 )
                        /*** Ein Errorcounter hat sein Limit erreicht */
			result |= CAN_ERR_BUSERROR ;
        }

        if ( ! received )
        	result |= CAN_ERR_RCVEMPTY ;

        return result ;
}


/************************************************************************
 *  CAN_Write()
 *  Message mit einer bestimmten Kennung auf den CAN-Bus schicken.
 *  Die Message aus 'PCANMsg->' uebernehmen.
 *  Algorithmus angelehnt an [AMC], S. 12
 *
 *	Parameter:
 *      PCANMsg: Zeiger auf Messagepuffer fuer die Rueckgabe der
 *		empfangenen Message.
 *  Rueckgabe: Bitmaske mit Fehlern
 *		0: Message wurde erfolgreich abgesetzt.
 *		CAN_ERR_XMTFULL: Es ist noch eine Uebertragung aktiv,
 *			Message wurde NICHT abgesetzt.
 */
int     CAN_Write(TCANMsg *PCANMsg) {
        unsigned char tmp ;

        /* Check 'Transmit Buffer Access'  */
        if ( ! (CANREAD(STATUS) & 0x04) )
                return CAN_ERR_XMTFULL ;      /* transmit buffer locked */
        if ( CANREAD(STATUS) & 0x20 )
                return CAN_ERR_XMTFULL ;

        /* Transmit Puffer fuellen */
        CANWRITE(XMTID, (unsigned char)(PCANMsg->ID >> 3) ) ;
        /* montiere das Laengenbyte der Message */
        tmp = ( ((unsigned char) PCANMsg->ID & 0x07) << 5) | (PCANMsg->LEN & 0x0f) ;
        if ( PCANMsg->RTR )
            /* sende remote-request-Message */
            CANWRITE( XMTLEN, tmp | 0x10 ) ;
        else {
            /* sende Datenmessage */
            CANWRITE( XMTLEN, tmp ) ;

            /* den ganzen Sendepuffer in den Controller bewegen */
            CANWRITE(XMTBUF+0, PCANMsg->DATA[0]) ;
            CANWRITE(XMTBUF+1, PCANMsg->DATA[1]) ;
            CANWRITE(XMTBUF+2, PCANMsg->DATA[2]) ;
            CANWRITE(XMTBUF+3, PCANMsg->DATA[3]) ;
            CANWRITE(XMTBUF+4, PCANMsg->DATA[4]) ;
            CANWRITE(XMTBUF+5, PCANMsg->DATA[5]) ;
            CANWRITE(XMTBUF+6, PCANMsg->DATA[6]) ;
            CANWRITE(XMTBUF+7, PCANMsg->DATA[7]) ;
        }
        /* transmit data by setting of TRANSMISSION REQUEST */
        CANBITSET(COMMAND, 0x01) ;

        return CAN_ERR_OK ;
}


/*******************************************************************
 *  CAN_Status()
 *	Abfrage des Statusregisters des 82C200 und Rueckgabe
 *	einer oder mehrere CAN_ERR_xxx Bedingungen.
 */
int     CAN_Status(void) {
	unsigned char tmp ;
        int	result = 0 ;

        tmp = CANREAD(STATUS) ;

        if ( ! (tmp & 0x01) )
        	result |= CAN_ERR_RCVEMPTY ;
        if ( tmp & 0x02 )
               	result |= CAN_ERR_OVERRUN ;
        if ( ! (tmp & 0x04) )
                result |= CAN_ERR_XMTFULL ;      /* transmit buffer locked */
        if ( tmp & 0x20 )
                result |= CAN_ERR_XMTFULL ;
       	if ( tmp & 0x40 )
		result |= CAN_ERR_BUSERROR;
        if ( tmp & 0x80 )
        	result |= CAN_ERR_BUSOFF ;


        return result ;
}


/*******************************************************************
 *  CAN_Init()
 *  Wird nach dem Anschalten einmal durchlaufen,
 *  und bei Re-Initialisierung.
 *	Fuehrt Registertest auf dem 82C200 aus,
 *  programmiert dann Baudrate, Konfiguration der Sende-/Empfangstreiber.
 *  Die Acceptance-Mask wird fuer den Empfang aller Messages gesetzt.
 *	
 *  Parameter:
 *		PortBase: Basisadresse der installierten Hardware.
 *		btr0btr1: Wert fuer die Timingregister des 82C200.
 */
int    CAN_Init(int _portbase, unsigned short int btr0btr1)
{
	int  		i ;

        PortBase = _portbase ;
        CANHWINIT(PortBase) ;

        /* RESET, no Interrupts, Pin REF is Output,
         * Synch only by transitions from recessive to dominant  */
        CANWRITE(CONTROL, 0x01) ;

        /* Register 4..8 sind 8 Bit read/write; 20..29 ist XMT-Buffer */
        for (i=4 ; i < 30 ; i++) if ( i < 9 || i > 19) {
                CANWRITE(i, i+5) ; // jedes Register mit 'Adresse+5' testen
                if ( CANREAD(i) != i+5 ) return CAN_ERR_REGTEST ;
        }
        for (i=4 ; i < 30 ; i++) if ( i < 9 || i > 19)
                if ( CANREAD(i) != i+5 ) return CAN_ERR_REGTEST ;



        CANWRITE(COMMAND, 0x0E) ; /* CLr Overrun, Release RcvBuff, Abort Xmission */

        /* Acceptance Code = Kennung fuer Empfang */
        CANWRITE(ACC_CODE, 0) ;
        CANWRITE(ACC_MASK, 0xff) ;   /* Kein Bit der Kennung ist relevant */

        nomsgregistered = 1 ;        /* Flag fuer  'CAN_registerMsg()' */

        /* TIMING0/1 fuer versch. ISO-Baudraten bei 16Mhz Quartz
         * [PLC] S.4, ausfuehrlich [TDS] S. 46f
         */
        CANWRITE(TIMING0, (btr0btr1 >> 8) & 0xFF) ;
        CANWRITE(TIMING1, btr0btr1 & 0xFF) ;

        /* Output Control: Push/Pull an TX0, TX1 disabled
         * OCTP1 = OCTN1 = OCPOL1 = 0,
         * OCTP0 = OCTN0 = 1, OCPOL0 = 0 (keine Invertierung)
         * OCMODE1 = 1, OCMODE0 = 0  (normal)
         * [TDS] S.50 u. 119 */
        CANWRITE(OUTPUTCTRL, 0 | 0x18 | 0x02) ;

        CANWRITE(CONTROL, 0x00) ; /* Clear reset-request */

        controlleractive = 1 ;

        return CAN_ERR_OK ;
}


/*******************************************************************
 *  CAN_registerMsg()
 *  programmiert das Empfangsfilter so um, dass zusaetzlich zu allen
 *  anderen Messages jetzt auch noch 'ID' empfangen wird.
 *
 *  - Beim ersten Aufruf von 'CAN_registerMsg()' wird vom Empfang
 *    aller Messages auf den Empfang eine einzigen umgeschaltet.
 *  - reset durch 'CAN_Init()'
 */
void    CAN_registerMsg(unsigned short int ID)
{
        unsigned char   n ;

        /* RESET, keine Interrupts, Pin REF ist Output,
         * Synch. nur bei Uebergaengen von Rezessiv nach Dominant  */
        CANWRITE(CONTROL, 0x01) ;

        /* Algorithmus:
         * Acceptance-Code bleibt immer die beim 1. Aufruf angegebene ID.
         * Folgende Aufrufe loeschen nur noch Bits in der Acceptance-Mask.
         */
        n = (ID >> 3) ;     /* nur die obersten 8 Bit der ID sind relevant */
        if ( nomsgregistered ) {
                nomsgregistered = 0 ;
                CANWRITE(ACC_CODE, n) ;
                CANWRITE(ACC_MASK, 0) ; /* Alle Bits der ID sind relevant */
        } else {
                /* Check, in welchen Bits 'n' vom Acceptance-Code differiert.
                 * Diese Bits zustzlich in der Acceptance-Mask setzen
                 * (XOR ist 1 bei Differenz) */
                n ^= CANREAD(ACC_CODE) ; /* n : high = differierende Bits */
                n |= CANREAD(ACC_MASK) ; /* n : high = neue 'dont care' Bits */
                CANWRITE(ACC_MASK, n) ;
        }

        CANWRITE(CONTROL, 0x00) ;      /* Reset Ende */
}

/*******************************************************************
 *  CAN_Close()
 *	Controller deaktivieren.
 *	Neustart dann mit CAN_Init() ;
 */
void CAN_Close(void) {
	if ( controlleractive )
		CANWRITE(CONTROL, 0x01) ; /* Controller in Reset-Mode */
        controlleractive = 0 ;
}

