
/**************************************************************************
     DONGLREG.C
     C-Modul zur Ansteuerung der
     Register des CAN-Controllers 82C200 auf dem
     PEAK-CAN-Dongle.

     Allgemeines zur Ansteuerung:
     ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
     Die Ansteuerung des PEAK-CAN-Dongles erfolgt ueber den LPT-Port.
     Die Signaladern des LPT-Portes sind ueber 3 Register im I/O-Bereich
     des PC's ansprechbar. Die Register liegen bei folgenden Adressen:
     "A" = Basisadresse      (Basisadr. z.B. 0x378 bei LPT1)
     "B" = Basisadresse+1
     "C" = Basisadresse+2

     Funktion der Registerbits:

     Signal          Richtung        Port
     --------        ---------       -------------------
     DATA0..DATA7    O (Basis-Mode)  "A"
                     I/O (Bidi-Mode)
     STROBE*         O               "C", Bit 0
     ALF*            O               "C", Bit 1
     ERR*            I               "B", Bit 3
     INI*            O               "C", Bit 2
     DSL*            O               "C", Bit 3
     ACK*            I               "B", Bit 6 (loest Interrupt aus)
     BUSY            I               "B", Bit 7
     PEND            I               "B", Bit 5
     SEL             I               "B", Bit 4

     INT_ENABLE      (intern)        "C", Bit 4
     READ_ENABLE     (intern)        "C", Bit 5 (nur BIDI-Schnittstellen)

     Manche Port-C-Leitungen werden zusaetzlich invertiert,
     daher muessen Ausgaben nach Port C immer mit einem "XOR 0x0B"
     verrechnet werden.


     Konventionelle und bidirektionale LPT-Ports:
     ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
     Im allgemeinen ist DATA0..7 des LPT-Port nicht bidirektional.
     Ausgaben an den CAN-Controllers gehen ueber DATA0..7,
     Ergebnisse werden in 2 4-Bit-Nibbles ueber Port "B" rueckgelesen.
     (sogenannter "Multiplex-Modus").

     Wurde im PC-BIOS der LPT-Port auf "EPP" oder "EPC" gestellt,
     kann der LPT-Port ueber das ECR-Register bei Basisadresse+0x402
     in den PS/2-Modus geschaltet werden.
     Dann koennen Daten schneller ueber DATA0..7 rueckgelesen werden.

     Allgemeines zum CAN-Controller:
     ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
     Der CAN-Controller 82C200 besitzt 32 interne Register.
     Diese sind 0..31 nummeriert.
     Ein Schreiben nach Register 31 loest einen Hardwarereset
     des 82C200 aus. Dies ist Harwareseitig durch das CPLD
     implementiert.


     Copyright (c) 1999 PEAK-System Technik GmbH, Darmstadt.
     ************************************************************/


/* Prototypen der definierte Routinen:
   Uebergabeparameter:
  'basisadr': Adresse des LPT-Ports (z.B. '0x378' fuer LPT1).
  'reg_adr':  anzusprechendes Register im 82C200
              (z.B. '2' fuer das Statusregister).
  'reg_val':  in ein Register zu schreibender Wert.
*/

/* Schreib-/Leseroutinen fuer den MUX-Modus */
void mux_reginit(int basisadr) ;
unsigned char mux_regread(int basisadr, unsigned char reg_adr) ;
void mux_regwrite(int basisadr, unsigned char reg_adr, unsigned char reg_val) ;

/* Schreib-/Leseroutinen fuer den EPP-Modus */
void epp_reginit(int basisadr) ;
unsigned char epp_regread(int basisadr, unsigned char reg_adr) ;
void epp_regwrite(int basisadr, unsigned char reg_adr, unsigned char reg_val) ;




/* Definitionen fuer Byte-weisen I/O-Zugriff */
#include <conio.h>
#define _outp outp
#define _inp inp

/* Tabelle, die den an PB3..PB7 gelesenen Bits
   das originale Nibble am Ausgang des Multiplexers zuordnet.
   Besonderheit: PB6 wird ignoriert.
 */
static unsigned char nibble_decode[32] = {
            0x8, 0x9, 0xa, 0xb, 0xc, 0xd, 0xe, 0xf,
            0x8, 0x9, 0xa, 0xb, 0xc, 0xd, 0xe, 0xf,
            0x0, 0x1, 0x2, 0x3, 0x4, 0x5, 0x6, 0x7,
            0x0, 0x1, 0x2, 0x3, 0x4, 0x5, 0x6, 0x7 } ;


void mux_reginit(int basisadr)
{
 /* Nicht definiert */
}

unsigned char mux_regread(int basisadr, unsigned char reg_adr)
{
        int PA = basisadr, PB = basisadr+1, PC = basisadr+2 ;
        unsigned char  b0, b1 ;

        /* Strobe (PC0) H->L: Adressuebernahme.
           Strobe L->H: Busfreigabe
           Mode (PC1) = L: MUX-Mode
           Nibbleumschaltung in der Lesephase mit PA6:
           PA6=0: PB3,PB4,PB5,PB7 = DATA4..7
           PA6=1:   "   "  "   "  = DATA0..3
              Bits in PC sind:   0 0 0 1 ispEN x Mode Strb Bemerkung}
                                         ~~~~~ ~ ~~~~ ~~~~ ~~~~~~~~~~~~~~
        */
        /* 1) Register-Adresse ausgeben */
        _outp(PC,0x0B ^ 0x1D );       /* H     H L    H    Grundstellung */
        _outp(PA,(reg_adr & 0x1F) | 0x80);/* Adresse. Bit 7 = 1 : Lesezyklus */
                                      /* Bit 6 = 0: zuerst High Nibble  */
        _outp(PC, 0x0B ^ 0x1C) ;      /* H     H L    L    Adr-Uebernahme mit Strb */
        /* 2) Daten lesen */
        b1=nibble_decode[_inp(PB)>>3];/* PA6=0: Nibble an PB = High Nibble */
        _outp(PA, 0x40) ;             /* PA6 = 1 */
        b0=nibble_decode[_inp(PB)>>3];/* PA6=1: Nibble an PB = Low Nibble */
        _outp(PC, 0x0B ^ 0x1D) ;      /* H     H L    H    Grundstellung  */

        return  (b1 << 4) | b0 ;
}

void mux_regwrite(int basisadr, unsigned char reg_adr, unsigned char reg_val)
{
        int PA = basisadr, PC = basisadr+2 ;

        /* Strobe (PC0) H->L: Adressuebernahme.
           Strobe L->H: Busfreigabe
           Mode (PC1) = L: MUX-Mode
              Bits in PC sind:   0 0 0 1 ispEN x Mode Strb Bemerkung
                                         ~~~~~ ~ ~~~~ ~~~~ ~~~~~~~~~~~~~~
        */
        /* 1) Register-Adresse ausgeben */
        _outp(PC, 0x0B ^0x1D) ;       /* H     H L    H    Grundstellung */
        _outp(PA, reg_adr & 0x1F) ;   /* Adresse. Bit 7 = 0 : Schreibzyklus */
        _outp(PC, 0x0B ^ 0x1C) ;      /* H     H L    L    Adr-Uebernahme */
        /* 2) Daten schreiben */
        _outp(PA, reg_val) ;          /* Schreibdaten ausgeben und ... */
        _outp(PC, 0x0B ^ 0x1D) ;      /* H     H L    H    ... uebernehmen */
                                      /* Und wieder in Grundstellung */
}


void epp_reginit(int basisadr)
{
        // 1284-kompatiblen Port in den "PS/2"-Mode setzen.
        // Damit das ECR-Register bei Offset 0x402 freigeschaltet ist,
        // mu im BIOS "EPP/EPC" eingestellt werden sein.
        _outp(basisadr+0x402, (_inp(basisadr+0x402) & 0x1F) | 0x20) ;
}


unsigned char epp_regread(int basisadr, unsigned char reg_adr)
{
        int PA = basisadr, PC = basisadr+2 ;
        unsigned char   wert ;

        /* Strobe (PC0) H->L: Adressuebernahme.
           Strobe L->H: Busfreigabe
           Mode (PC1) = H EPP-Mode }
              Bits in PC sind:   0 0 0 1 ispEN x Mode Strb Bemerkung
                                         ~~~~~ ~ ~~~~ ~~~~ ~~~~~~~~~~~~~~
        */
        /* 1) Register-Adresse ausgeben */
        _outp(PC, 0x0B ^ 0x1F) ;       /* H     H H    H    Grundstellung */
        _outp(PA, (reg_adr& 0x1F)|0x80) ; /* Adresse. Bit 7 = 1 : Lesezyklus */
        _outp(PC, 0x0B ^ 0x3E) ;       /* H     H H    L    Adr-Uebernahme, und */
                                       /* Datenport Tristate schalten mit Bit 5 */
        /* 2) Daten lesen */
        wert = _inp(PA) ;              /* CPLD muss Daten schnell genug anlegen! */
        _outp(PC, 0x0B ^ 0x1F) ;       /* H     H H    H    Grundstellung */

        return  wert ;
}

void epp_regwrite(int basisadr, unsigned char reg_adr, unsigned char reg_val)
{
        int PA = basisadr, PC = basisadr+2 ;

        /* Strobe (PC0) H->L: Adressuebernahme.
           Strobe L->H: Busfreigabe
           Mode (PC1) = L: MUX-Mode
              Bits in PC sind:   0 0 0 1 ispEN x Mode Strb Bemerkung
                                         ~~~~~ ~ ~~~~ ~~~~ ~~~~~~~~~~~~~~
        */
        /* 1) Register-Adresse ausgeben */
        _outp(PC, 0x0B ^0x1D) ;       /* H     H L    H    Grundstellung */
        _outp(PA, reg_adr & 0x1F) ;   /* Adresse. Bit 7 = 0 : Schreibzyklus */
        _outp(PC, 0x0B ^0x1E) ;       /* H     H L    L    Adr-Uebernahme */
        /* 2) Daten schreiben */
        _outp(PA, reg_val) ;          /* Schreibdaten ausgeben und ... */
        _outp(PC, 0x0B ^ 0x1F) ;      /* H     H L    H    ... uebernehmen */
                                      /* Und wieder in Grundstellung */
}
