

#define __CANC__
/**********************************************************************
 *      can.c
 *	Funktion:
 *	Treiberinterface fuer eine SJA1000-basierte CAN-Interfacehardware.
 *      Der SJA 1000 wird im PELICAN-Mode betrieben, CAN2.0B wird
 *      unterstuetzt.
 *
 *	Der Treiber arbeitet im Pollingverfahren, und gestattet das Lesen,
 *	Schreiben und Konfigurieren des CAN-Interfaces.
 *
 *      Bereitgestellte #defines:
 *	HARDWARETYP	    - Einstellen der verwendeten Hardware.
 *
 *      Bereitgestellte Prozeduren:
 *      CAN_Init()          - Modul initialisieren fuer Empfang aller Msgs
 *      CAN_registerMsg()   - Empfangsfilter fuer Msg erweitern.
 *      CAN_Read()          - Empfangspuffer lesen und Status zurueck,
 *				evtl. Controller reaktivieren.
 *      CAN_Write()         - Sendepuffer schreiben und Status zurueck
 *      CAN_Status()        - Statusregister des CAN-Controllers
 *      CAN_Close()         - Controller deaktivieren.
 *
 *      Autor   : J. Hoppe,      (c) 1999 PEAK-System Technik
 *      Sprache : Borland-C
 *
 *
 *      Revision-History:
 *      ----------------------------------------------------------------
 *      1.6.1996:       Portierung nach PC und 82C200.
 *      25.8.1998:      Umstellung auf SJA1000 / PELICAN
 *      ----------------------------------------------------------------
 *
 *      Literatur:
 *      ----------------------------------------------------------------
 *	[DS]	Data Sheet SJA1000, 4.11.1997, Philips
 *
 *
 *
 *      Copyright (c) 1999 PEAK-System Technik GmbH, Darmstadt.
 *      */

#include "can.h"        /* eigene Typendeklarationen */


/************************************************************************/
/***    Konstanten                                                    ***/
/************************************************************************/

/* Auswahl der anzusteuernden CAN-Hardware:
 * ISA-Einsteckkarte oder eine der Donglevarianten
 * 'PortBase' muss von Hauptprogramm richtig eingestellt werden! */

#undef ISACAN		/* nur eines dieser Symbole #definen ! */
#define DONGLEMUX
#undef DONGLEEPP

#ifdef ISACAN
/***** Ansteuerung der ISA-Karte: ISAREG.C einbinden! *****/
/* Prototypen der Treiberroutinen in ISAREG.C */
unsigned char isa_regread(unsigned short basisadr, unsigned char reg_adr) ;
void isa_regwrite(unsigned short basisadr, unsigned char reg_adr, unsigned char reg_val) ;
/* Ansteuerung definieren */
#define CANHWINIT(adr)
#define CANREAD(adr) isa_regread(PortBase, (unsigned char)(adr))
#define CANWRITE(adr,val) isa_regwrite( PortBase, (unsigned char)(adr), (unsigned char)(val))

#endif

#ifdef DONGLEMUX
/***** Ansteuerung des Dongles im MUX-Mode: DONGLEREG.C einbinden! *****/
/* Prototypen der MUX-Treiberroutinen in DOBGLEREG.C */
void mux_reginit(unsigned short basisadr) ;
unsigned char mux_regread(unsigned short basisadr, unsigned char reg_adr) ;
void mux_regwrite(unsigned short basisadr, unsigned char reg_adr, unsigned char reg_val) ;
/* Ansteuerung definieren */
#define CANHWINIT     mux_reginit(PortBase)
#define CANREAD(adr) mux_regread(PortBase, (unsigned char)(adr))
#define CANWRITE(adr,val) mux_regwrite( PortBase, (unsigned char)(adr), (unsigned char)(val))

#endif

#ifdef DONGLEEPP
/***** Ansteuerung des Dongles im EPP-Mode: DONGLEREG.C einbinden! *****/
/* Prototypen der EPP-Treiberroutinen in DONGLEREG.C */
void epp_reginit(unsigned short basisadr) ;
unsigned char epp_regread(unsigned short basisadr, unsigned char reg_adr) ;
void epp_regwrite(unsigned short basisadr, unsigned char reg_adr, unsigned char reg_val) ;
/* Ansteuerung definieren */
#define CANHWINIT epp_reginit(PortBase)
#define CANREAD(adr) epp_regread(PortBase, (unsigned char)(adr))
#define CANWRITE(adr,val) epp_regwrite( PortBase, (unsigned char)(adr), (unsigned char)(val))


#endif

/*** Adressen im SJA 1000 CAN-Controller, PeliCAN-Mode ***/

#define REG_MODE            0
#define REG_COMMAND         1
#define REG_STATUS          2
#define REG_INTERRUPT       3
#define REG_IRQENABLE       4           // Interrupt Enable
#define REG_TIMING0         6           // Bus Timing 0
#define REG_TIMING1         7           // Bus Timing 1
#define REG_OUTPUTCTRL      8           // Output Control
#define REG_TEST            9

#define REG_ARBLOSTCAP     11           // Arbitration Lost Capture
#define REG_ERRCODECAP     12           // Error Code Capture
#define REG_ERRWARNLIMIT   13           // Error Warning Limit
#define REG_RXERRCOUNT     14           // Rx Error Counter
#define REG_TXERRCOUNT     15           // Tx Error Counter

/**** Rcv/Xmt-Buffer only in operation mode ****/
// read = Rx-Buffer, write = TxBuffer
#define REG_FRAMEINFO      16           // Frame format, RTR, data length
// SFF-Mode (standard frames)
#define REG_SFFID1         17           // bits 28..21 of ID
#define REG_SFFID2         18           // bits 20..18 of ID
#define REG_SFFDATA        19           // Begin of data buffer (8 bytes)
// EFF-Mode (extended frames)
#define REG_EFFID1         17           // bits 28..21 of ID
#define REG_EFFID2         18           // bits 20..13 of ID
#define REG_EFFID3         19           // bits 12..5 of ID
#define REG_EFFID4         20           // bits 4..0 of ID
#define REG_EFFDATA        21           // Begin of data buffer (8 bytes)

/**** Acceptance filter only in reset mode ****/
#define REG_ACCCODE0       16
#define REG_ACCCODE1       17
#define REG_ACCCODE2       18
#define REG_ACCCODE3       19
#define REG_ACCMASK0       20
#define REG_ACCMASK1       21
#define REG_ACCMASK2       22
#define REG_ACCMASK3       23

#define REG_RCVBUFSTART    30           // receive buffer start address
#define REG_CLOCKDIV       31           // Clock divider





/************************************************************************/
/***    Lokale Variable                                               ***/
/************************************************************************/

static unsigned short	PortBase ;    /* Basisadresse der Hardware aus CAN_Init() */
static int 		nomsgregistered ;  /* Flag fuer  'CAN_registerMsg()' */
static unsigned long	TheAccCode, TheAccMask ;
static unsigned short	TheBtr0Btr1 ;

/************************************************************************/
/***    Lokale Prozeduren dieses Moduls                               ***/
/************************************************************************/

/**************************************************************************
 *      isinstalled()
 *      Pruefung, ob der SJA 1000 zugreifbar ist.
 *      Wird dazu resetet, und ein Registertest wird durchgefuehrt.
 *      Return:
 *              0 = Registertest fehlgeschlagen
 *              1 = OK
 *      */
int     isinstalled(void)
{
        int     i, n ;
        CANHWINIT ;
        CANWRITE(REG_MODE, CANREAD(REG_MODE) | 0x01 ) ; /* set RESET */

        // set PeliCAN-Mode, leave Clockout und CBP as set by HARDRESET
        CANWRITE(REG_CLOCKDIV, CANREAD(REG_CLOCKDIV) | 0x80 ) ;
#ifdef DEBUGOUT
if ( CANREAD(REG_CLOCKDIV) & 0x80 != 0x80 )
	printf("Chip nicht im PeliCAN-Mode!") ;
printf("Registertest 1 ...") ;
         // Register 6..8,13,16..23 sind 8 Bit read/write
         // (Register 14,15 werden durch Bus-Off veraendert!)
#endif
        for (i=6 ; i <= 23 ; i++) if ( i <= 8 || i == 13 || i >= 16) {
                CANWRITE(i, i+5) ; // jedes Register mit 'Adresse+5' testen
                n = CANREAD(i) ;
                if ( n != i+5 ) {
#ifdef DEBUGOUT
printf("Fehler bei Registertest 1: Reg[%d] ist 0x%x, soll sein 0x%x", i, n, i+5) ;
#endif
                        return 0 ;
                }
        }
#ifdef DEBUGOUT
printf("Registertest 2 ...") ;
#endif
        for (i=6 ; i <= 23 ; i++) if ( i <= 8 || i == 13 || i >= 16) {
                n = CANREAD(i) ;
                if ( n != i+5 ) {
#ifdef DEBUGOUT
printf("Fehler bei Registertest 2: Reg[%d] ist 0x%x, soll sein 0x%x", i, n, i+5) ;
#endif
                        return 0 ;
                }
        }
        return 1 ;
}




/**********************************************************************
 *      activate()
 *      Controller, wenn moeglich, hardresetten,
 *      neu programmieren, und aktiv schalten.
 *      */
int     activate(
        unsigned short int btr0btr1,
        unsigned long int acc_code, // code+mask auch 32Bit, wie ID!
        unsigned long int acc_mask )
{

//        HARDRESET ;

        // reset
//        CANWRITE(REG_MODE, CANREAD(REG_MODE) | 0x01 ) ;

        // set PeliCAN-Mode, leave Clockout und CBP as set by HARDRESET
//        CANWRITE(REG_CLOCKDIV, CANREAD(REG_CLOCKDIV) | 0x80 ) ;

	// HardReset, goto PeliCAN, and Registertest
	if ( ! isinstalled() )
		return 	CAN_ERR_REGTEST ;

        // no sleep (MOD.4==0),
        // vorlaeufig single acceptance filter (MOD.3==1),
        // no self test mode (MOD.2=0)
        // no listen only(MOD.1==0)
        // reset (MOD.0==1)
        CANWRITE(REG_MODE, 0x09) ;


        //  Clr Overrun, Release Rcvbuf, Abort Xmission
        CANWRITE(REG_COMMAND, 0x0E ) ;

        // Acceptance Code = Kennung fuer Empfang.
        // gueltig sind Bits 28..0, diese muessen auf
        // ACR/AMR0..3 verteilt werden.
        // keine Unterstuetzung des RTR-Bits, keine Unterstuetzung
        // der data bytes 0 und 1 bei standard frames-filterung
        // -> wird in registerMessages() unterdrueckt.
//sprintf(diagmsg, "acc_code=%x, acc_mask=%x", acc_code, acc_mask) ; putdiagmsg() ;
        CANWRITE(REG_ACCCODE0,  (acc_code >> 21) & 0xFF) ;
        CANWRITE(REG_ACCCODE1,  (acc_code >> 13) & 0xFF) ;
        CANWRITE(REG_ACCCODE2,  (acc_code >>  5) & 0xFF) ;
        CANWRITE(REG_ACCCODE3, ((acc_code << 3) & 0xFF) | 0x07) ;
        CANWRITE(REG_ACCMASK0,  (acc_mask >> 21) & 0xFF) ;
        CANWRITE(REG_ACCMASK1,  (acc_mask >> 13) & 0xFF) ;
        CANWRITE(REG_ACCMASK2,  (acc_mask >>  5) & 0xFF) ;
        CANWRITE(REG_ACCMASK3, ((acc_mask << 3) & 0xFF) | 0x07) ;

        /* Baudratenregister setzen */
        CANWRITE(REG_TIMING0, (btr0btr1 >> 8) & 0xFF ) ;
        CANWRITE(REG_TIMING1, btr0btr1 & 0xFF) ;

        /* Output Control: Push/Pull an TX0, TX1 disabled
         * OCTP1 = OCTN1 = OCPOL1 = 0,
         * OCTP0 = OCTN0 = 1, OCPOL0 = 0 (keine Invertierung)
         * OCMODE1 = 1, OCMODE0 = 0  (normal)
         * [DS] S.50 u. 119 */
        CANWRITE(REG_OUTPUTCTRL, 0 | 0x18 | 0x02) ;

        /* Clear Interruptflags */
        CANREAD(REG_INTERRUPT) ;

        // disable all Interrupts
        CANWRITE(REG_IRQENABLE, 0x00) ;

        /* Reset Ende */
        CANWRITE(REG_MODE, 0x08) ;
	//wg. err-test     CANWRITE(REG_MODE, CANREAD(REG_MODE) & ~0x01 ) ;

        return CAN_ERR_OK ;
}




/************************************************************************/
/***    Global verwendbare Prozeduren (Methoden) dieses Moduls        ***/
/************************************************************************/


/************************************************************************
 *      CAN_Read()
 *      Polling-Routine:
 *      -  empfangene Message in den Puffer 'PCANMsg->' uebertragen
 *      -  Ueberwachung des Bus-Status und ggf. Reset des CAN-Controllers.
 *
 *	Parameter:
 *               PCANMsg: Zeiger auf Messagepuffer fuer die Rueckgabe der
 *		 empfangenen Message.
 *
 *      Rueckgabe: Bitmaske mit Fehlern.
 *		0 :	Kein Fehler, neue Message steht in 'PCANMsg->'.
 *		CAN_ERR_RCVEMPTY: es wurde keine neue Message empfangen.
 *		CAN_ERR_OVERRUN : eine empfangene Message wurde nicht rechtzeitig
 				ausgelesen.
 *		CAN_ERR_BUSERROR: Controller meldet viele Busstoerungen.
 *		CAN_ERR_BUSOFF: Controller war Bus-Off und wurde wieder
 *			aktiviert.
 *      */
int     CAN_Read(TCANMsg *PCANMsg)
{
        int     received = 0 ;
        int	result = 0 ;
        int     i, len ;
        int     frameinfo ;

        /************************ Datenempfang **************************/
	if ( CANREAD(REG_STATUS) & 0x01 ) {
		received = 1 ;
		frameinfo = CANREAD(REG_FRAMEINFO) ;
		PCANMsg->LEN = len = frameinfo & 0x0f ;
		if ( len > 8 ) len = 8 ; /* Laenge ggf. korrigieren */
		if ( frameinfo & 0x80 ) {
			unsigned long tmp ;
			/*** extended frame ***/
			PCANMsg->EXTENDED = 1 ;
			tmp =   ( (unsigned long) CANREAD(REG_EFFID1) << 24)
			      | ( (unsigned long) CANREAD(REG_EFFID2) << 16)
			      | ( (unsigned long) CANREAD(REG_EFFID3) <<  8)
			      | ( (unsigned long) CANREAD(REG_EFFID4)      )  ;
			PCANMsg->ID = tmp >> 3 ;
			if ( frameinfo & 0x40 )
				PCANMsg->RTR = 1 ;
			else	{
				PCANMsg->RTR = 0 ;
				/* kein RTR: den ganzen Empfangspuffer nach 'data[]' bewegen */
				for (i=0 ; i < len ; i++)
				         PCANMsg->DATA[i] = CANREAD(REG_EFFDATA+i) ;
			}
		} else {
			/*** standard frame, wie 82C200 ***/
			PCANMsg->EXTENDED = 0 ;
			PCANMsg->ID = ((unsigned short) CANREAD(REG_SFFID1) << 3) | (CANREAD(REG_SFFID2) >> 5) ;

			if ( frameinfo & 0x40 )
				PCANMsg->RTR = 1 ;
			else	{
				PCANMsg->RTR = 0 ;
				/* kein RTR: den ganzen Empfangspuffer nach 'data[]' bewegen */
				for (i=0 ; i < len ; i++)
				         PCANMsg->DATA[i] = CANREAD(REG_SFFDATA+i) ;
			}
		}
//sprintf(diagmsg, "pelican-rcv: frameinfo=%x, msgtype=%x, len=%d, id=%x",
//frameinfo, *msgtype, *len, *id) ; putdiagmsg() ;

		/* release receive buffer */
	        CANWRITE(REG_COMMAND, 0x04) ;
	}

	/********************** Evtl. 'Overrun'-Bedingung loeschen *****/
	if ( CANREAD(REG_STATUS) & 0x02 ) {
		result |= CAN_ERR_OVERRUN ;
		CANWRITE(REG_COMMAND, 0x08) ;
	}


        /********************* Fehlerverarbeitung ***********************/
        if ( CANREAD(REG_STATUS) & 0x80 ) {
            /* Bus-Off: Controller is NOT involved in bus activities */
            result |= CAN_ERR_BUSOFF ;
            /* Reset request auf 'absent' = Bit 0 loeschen */
            CANWRITE(REG_MODE, CANREAD(REG_MODE) & ~0x01) ;
            /* Der Controller geht wieder 'On-Bus' nach dem Empfang
             * von 128 * 11 aufeinanderfolgenden rezessiven Bits.
             */
        } else {
            /* Bus-On: Controller is involved in bus activities */
            if ( CANREAD(REG_STATUS) & 0x40 )
                /* schwere Stoerung: ein Errorcounter erreichte Limit */
                result |= CAN_ERR_BUSERROR ;
	}

	if ( ! received )
		result |= CAN_ERR_RCVEMPTY ;

	return result ;
}


/************************************************************************
 *      CAN_Write()
 *      Message mit einer bestimmten Kennung auf den CAN-Bus schicken.
 *      Die Message aus 'PCANMsg->' uebernehmen.
 *
 *	Parameter:
 *               PCANMsg: Zeiger auf Messagepuffer fuer die Rueckgabe der
 *		 empfangenen Message.
 *      Rueckgabe: Bitmaske mit Fehlern
 *		0: Message wurde erfolgreich abgesetzt.
 *		CAN_ERR_XMTFULL: Es ist noch eine Uebertragung aktiv,
 *			Message wurde NICHT abgesetzt.
 *      */
int     CAN_Write(TCANMsg *PCANMsg)
{
        int     tmp ;
        int     i, len ;

        /* Check 'Transmit Buffer Access'  */
        if ( ! (CANREAD(REG_STATUS) & 0x04) )
                return CAN_ERR_XMTFULL ;      /* transmit buffer locked */
        if ( CANREAD(REG_STATUS) & 0x20 )
                return CAN_ERR_XMTFULL ;


        tmp =  (PCANMsg->EXTENDED ? 0x80 : 0)
           | (PCANMsg->RTR        ? 0x40 : 0)
           | (PCANMsg->LEN & 0x0f) ;
        CANWRITE(REG_FRAMEINFO, tmp) ;
	len = PCANMsg->LEN ;
        if ( len > 8) len = 8 ;
// sprintf(diagmsg, "pelican-xmt: frameinfo=%x, msgtype=%x, len=%d, id=%x",
// tmp, msgtype, len, id) ; putdiagmsg() ;

        if (PCANMsg->EXTENDED) {
                /*** extended frame ***/
                CANWRITE(REG_EFFID1, PCANMsg->ID >> 21) ;
                CANWRITE(REG_EFFID2, PCANMsg->ID >> 13) ;
                CANWRITE(REG_EFFID3, PCANMsg->ID >> 5) ;
                tmp = (PCANMsg->ID & 0x1F) << 3 ;
                if ( PCANMsg->RTR )
                        /* setze sicherheitshalber hier noch remote-request */
                        CANWRITE(REG_EFFID4, tmp | 0x04 ) ;
                else {
                        /* sende Datenmessage */
                        CANWRITE(REG_EFFID4, tmp ) ;

                        /* den ganzen Sendepuffer in den Controller bewegen */
                        for (i=0 ; i < len ; i++)
                                CANWRITE(REG_EFFDATA+i, PCANMsg->DATA[i]) ;
               }
          } else {
                /*** standard frame ***/
                CANWRITE(REG_SFFID1, PCANMsg->ID >> 3 ) ;
                tmp = (PCANMsg->ID & 0x07) << 5 ;
                if ( PCANMsg->RTR )
                        /* setze sicherheitshalber hier noch remote-request */
                        CANWRITE(REG_SFFID2, tmp | 0x10 ) ;
                else {
                        /* sende Datenmessage */
                        CANWRITE(REG_SFFID2, tmp ) ;

                        /* den ganzen Sendepuffer in den Controller bewegen */
                        for (i=0 ; i < len ; i++)
                                CANWRITE(REG_SFFDATA+i, PCANMsg->DATA[i]) ;
               }
        }
         /* transmit data by setting of TRANSMISSION REQUEST */
        CANWRITE(REG_COMMAND, 0x01) ;

        return CAN_ERR_OK ;
}






/*******************************************************************
 *      CAN_Status()
 *	Abfrage des Statusregisters des 82C200 und Rueckgabe
 *	einer oder mehrere CAN_ERR_xxx Bedingungen.
 *	*/
int     CAN_Status(void) {
	unsigned char tmp ;
        int	result = 0 ;

        tmp = CANREAD(REG_STATUS) ;

        if ( ! (tmp & 0x01) )
        	result |= CAN_ERR_RCVEMPTY ;
        if ( tmp & 0x02 )
               	result |= CAN_ERR_OVERRUN ;
        if ( ! (tmp & 0x04) )
                result |= CAN_ERR_XMTFULL ;      /* transmit buffer locked */
        if ( tmp & 0x20 )
                result |= CAN_ERR_XMTFULL ;
       	if ( tmp & 0x40 )
		result |= CAN_ERR_BUSERROR;
        if ( tmp & 0x80 )
        	result |= CAN_ERR_BUSOFF ;


        return result ;
}


/*******************************************************************
 *      CAN_Init()
 *      Wird nach dem Anschalten einmal durchlaufen,
 *      und bei Re-Initialisierung.
 *	Schaltet SJA10000 in PeliCAN-Mode,
 *	fuehrt Registertest auf dem 82C200 aus,
 *      programmiert dann Baudrate, Konfiguration der Sende/Empfangstreiber.
 *      Die Acceptance-Mask wird fuer den Empfang aller Messages gesetzt.
 *	Parameter:
 *		PortBase: Basisadresse der installierten Hadrware.
 *		btr0btr1: Wert fuer die Timingregister des 82C200.
 *      */
int    CAN_Init(unsigned short _portbase, unsigned short int btr0btr1)
{
        PortBase = _portbase ; // als erstes, damit HWINIT,CANREAD,CANWRITE funktionieren.
	TheBtr0Btr1 = btr0btr1 ;
        CANHWINIT ;

	// SJA testen und initalisieren, mit (zunaechst) offenem Akzeptanzfilter
	TheAccCode = 0xffffffff ;
	TheAccMask = 0xffffffff ;

	return activate(btr0btr1, TheAccCode, TheAccMask) ;
}

/*******************************************************************
 *      CAN_registerMsg()
 *      programmiert das Empfangsfilter so um, da zustzlich zu allen
 *      anderen Messages jetzt auch noch die 'ID1' bis 'ID2' empfangen werden.
 *	Die Bitmuster-Bildung findet erst auf TheAccCode/TheAccMask statt,
 *	dann wird der SJA1000 neu initialisiert.
 *	'extended' gibt an, ob ID1/ID2 als 11 oder als 29Bit-ID gemeint sind.
 *	Intern werdenalle ID's als extended dargestellt, standard 11-Bit IDs werden nach
 *	CAN2.0B in 29Bit-IDs gewandelt.
 *
 *      - Beim ersten Aufruf von 'CAN_registerMsg()' wird vom Empfang
 *        aller Messages auf den Empfang eine einzigen umgeschaltet.
 *      - reset durch 'CAN_Init()'
 *      */
void    CAN_registerMsg(unsigned long ID1, unsigned long ID2, unsigned char extended)
{
	unsigned long	tmp, eqmask ;

            // Client-Filter erweitern

         // make standard to extended ID
	if ( ! extended ) { ID1 <<= 18 ; ID2 <<=18 ; }

        // welche Bits bleiben im Bereich msg1.ID..msg2.ID konstant?
        tmp = ID1 ^ ID2 ; // differierende Bits = 1
        // alle kleineren Bits hinter dem hoechsten differierenden Bit
        // differieren im Bereich msg1..msg2 auch irgendwann einmal
        eqmask = 0 ;
        while (tmp) {
		tmp >>= 1 ; // solange schieben, bis 0
                eqmask = (eqmask << 1) | 1 ;  // pro shift eine 1 mehr
        }
        if ( TheAccMask == 0xffffffff && TheAccCode == 0xffffffff) {
                // 1. Message
                TheAccMask = eqmask ;
                TheAccCode = ID1 ;
        } else {
                // Check, in welchen Bits 'n' vom Acceptance-Code differiert.
                // Diese Bits zusaetzlich in der Acceptance-Mask setzen
                // (XOR ist 1 bei Differenz, in C: "^")
                tmp = ID1 ^ TheAccCode ; // high = differierende Bits
                TheAccMask |= tmp ;      // high = neue 'dont care' Bits
                TheAccMask |= eqmask ; // diese Bits differieren im Bereich ID1..ID2 sowieso

                TheAccCode = ID1 ;
        }
        // Wenn ein Standard frame registriert:
        // registriere alle extended Messages mit diesen Bits 28..18,
        // -> Bits 17..0 = don't care.
        if ( ! extended )
               TheAccMask |= 0x3ffff ;

	activate(TheBtr0Btr1, TheAccCode, TheAccMask) ;
}

/*******************************************************************
 *      CAN_Close()
 *	Controller deaktivieren.
 *	Neustart dann mit CAN_Init() ;
 *	*/
void CAN_Close(void) {
	// SJA in Reset-Mode
        CANWRITE(REG_MODE, CANREAD(REG_MODE) | 0x01 ) ;
}



/***********************************************************************
 *      regname()
 *      gibt den Klartextnamen einer Registeradresse zurueck.
 *      */
static char * regname(int reg) {
        switch ( reg ) {
        case REG_MODE       : return "MODE" ;
        case REG_COMMAND    : return "COMMAND" ;
        case REG_STATUS     : return "STATUS" ;
        case REG_INTERRUPT  : return "INTERRUPT" ;
        case REG_TIMING0    : return "TIMING0" ;
        case REG_TIMING1    : return "TIMING1" ;
        case REG_OUTPUTCTRL : return "OUTPUTCTRL" ;

        default          : return "UNKNOWN" ;
        }
}


