UNIT can ;

  {***********************************************************************
   *      can.pas
   *      Funktion:
   *      Treiberinterface fuer eine 82C200-basierte CAN-Interfacehardware.
   *      Der Treiber arbeitet im Pollingverfahren, und gestattet das Lesen,
   *      Schreiben und Konfigurieren des CAN-Interfaces.
   *
   *      Bereitgestellte $defines:
   *      ISACAN,DONGLEMUX,DONGLEEPP - Einstellen der verwendeten Hardware.
   *
   *      Bereitgestellte Prozeduren:
   *      CAN_Init()          - Modul initialisieren fuer Empfang aller Msgs
   *      CAN_registerMsg()   - Empfangsfilter fuer Msg erweitern.
   *      CAN_Read()          - Empfangspuffer lesen und Status zurueck,
   *                              evtl. Controller reaktivieren.
   *      CAN_Write()         - Sendepuffer schreiben und Status zurueck
   *      CAN_Status()        - Statusregister des CAN-Controllers
   *      CAN_Close()         - Controller deaktivieren.
   *
   *      Autor   : J. Hoppe,     (c) 1999 PEAK-System TechnikGmbH, Darsmtadt.
   *      Sprache : Borland-Pascal
   *
   *
   *      Revision-History:
   *      ----------------------------------------------------------------
   *      20.7.1996      Portierung aus Borland-C nach Borland-Pascal
   *      ----------------------------------------------------------------
   *
   *      Literatur:
   *      ----------------------------------------------------------------
   *      [AMC]   Application of the P8xC592 microcontroller with
   *              CAN-interface.
   *              P. Buehring, P. Hank, Philips Application Note, 7/92
   *      [PLC]   CAN Physical Layer Concepts for the P8xC592.
   *              H. Eisele, Philips Application Note, 7/92
   *      [TDS]   80C592 Target Device Specification 1991 Vers. 2.1.
   *              Philips, August 1991
   *
   *
   *      Copyright     (c) 1999 PEAK-System Technik GmbH, Darmstadt.
   *      *}

  {* Auswahl der anzusteuernden CAN-Hardware: ISA-Einsteckkarte oder eine
   * der Donglevarianten.
   * !!! Nur eines dieser Symbole darf define'd sein !!!
   *}

  {$undef  ISACAN}        { define, wenn ISA-Einsteckkarte }
  {$define DONGLEMUX}     { define, wenn Dongle im Standard-Modus }
  {$undef  DONGLEEPP}     { define, wenn Dongle im EPP-Modus }

  INTERFACE

    USES
      {* Einbinden aller Hardware-Ansteuerungs-Unit's, aber nur eine wird
       * wirklich bentigt. }
       donglreg ;


    CONST
      {*** Konstantendefinitionen ***}

      {*** Baudratencodes = Registerwerte BTR0/BTR1 *}
      CAN_BAUD_1M   = $0014 ; {*   1 MBit / s *}
      CAN_BAUD_500K = $001C ; {* 500 kBit / s *}
      CAN_BAUD_250K = $011C ; {* 250 kBit / s *}
      CAN_BAUD_125K = $031C ; {* 125 kBit / s *}
      CAN_BAUD_100K = $432F ; {* 100 kBit / s *}
      CAN_BAUD_50K  = $472F ; {*  50 kBit / s *}
      CAN_BAUD_20K  = $532F ; {*  20 kBit / s *}
      CAN_BAUD_10K  = $672F ; {*  10 kBit / s *}
      CAN_BAUD_5K   = $7F7F ; {*   5 kBit / s *}


      {*** Maskenbits fuer Fehlerbedingungen ***}
      CAN_ERR_OK       = $00 ;
      CAN_ERR_XMTFULL  = $01 ; {* Sendepuffer im Controller ist voll *}
      CAN_ERR_RCVEMPTY = $02 ; {* keine neue Message empfangen *}
      CAN_ERR_OVERRUN  = $04 ; {* CAN-Controller wurde zu spaet gelesen *}
      CAN_ERR_BUSERROR = $08 ; {* Busfehler: ein Errorcounter erreichte Limit *}
      CAN_ERR_BUSOFF   = $10 ; {* Busfehler: CAN-Controller ging 'Bus-Off' *}
      CAN_ERR_REGTEST  = $20 ; {* Fehler beim Registertest des CAN-Controllers *}


    TYPE
      {*** Typendefinitionen ***}

      {* eine CAN-Message *}
      TCANMsg = RECORD
          ID  : word ;     {* 11 Bit-Kennung                         *}
          RTR : boolean ;  {* true, wenn remote request              *}
          LEN : byte ;     {* Anzahl der gueltigen Datenbytes (0..8) *}
          DATA: ARRAY[0..7] OF byte ; {* Datenbytes 0..7             *}
        END ;


    FUNCTION  CAN_Init(_portbase: word ; btr0btr1: word) : integer ;
    PROCEDURE CAN_registerMsg(ID: word) ;
    FUNCTION  CAN_Read (VAR CANMsg : TCANMsg) : integer ;
    FUNCTION  CAN_Write(VAR CANMsg : TCANMsg ) : integer ;
    FUNCTION  CAN_Status : integer ;
    PROCEDURE CAN_Close ;


  IMPLEMENTATION

    {************************************************************************}
    {***    Konstanten                                                    ***}
    {************************************************************************}

    CONST
      {**** Registeradressen des CAN_Controllers 82C200 ***}
      _CONTROL    =  0 ;
      _COMMAND    =  1 ;
      _STATUS     =  2 ;
      _INTERRUPT  =  3 ;
      _ACC_CODE   =  4 ;     {* Acceptance Code *}
      _ACC_MASK   =  5 ;     {* Acceptance Mask *}
      _TIMING0    =  6 ;     {* Bus Timing 0 *}
      _TIMING1    =  7 ;     {* Bus Timing 1 *}
      _OUTPUTCTRL =  8 ;     {* Output Control *}

      _XMTID      = 10 ;     {* Transmit Buffer: Identifier       *}
      _XMTLEN     = 11 ;     {*     RTR Bit und Data length Code  *}
      _XMTBUF     = 12 ;     {*     BYTE1 des data field's        *}

      _RCVID      = 20 ;     {* Receive Buffer: Identifier        *}
      _RCVLEN     = 21 ;     {*     RTR Bit und Data length Code  *}
      _RCVBUF     = 22 ;     {*     BYTE1 des data field's        *}


    VAR
      {************************************************************************}
      {***    Lokale Variable                                               ***}
      {************************************************************************}

      PortBase : word ;    {* Basisadresse der Hardwarem aus CAN_Init() *}
      nomsgregistered : boolean ;  {* Flag fuer  'CAN_registerMsg()' *}
      controlleractive : boolean ;


      {************************************************************************}
      {***    Lokale Prozeduren dieses Moduls                               ***}
      {************************************************************************}


      {*** Hardware initialisieren ***}
    PROCEDURE CANHWINIT(basisadr : word) ;
      BEGIN
        {* Auswahl der anzusteuernden CAN-Hardware: ISA-Einsteckkarte oder
         * eine der Donglevarianten) *}
        {$ifdef DONGLEMUX}  mux_reginit(basisadr) ; {$endif}
        {$ifdef DONGLEEPP}  epp_reginit(basisadr) ; {$endif}
      END { "CANHWINIT" } ;

    {*** CANREAD(): ein Register des 82C200 auslesen ***}
    FUNCTION CANREAD(reg_adr : byte) : byte ;
      BEGIN
        {* Auswahl der anzusteuernden CAN-Hardware: ISA-Einsteckkarte oder
         * eine der Donglevarianten *}
        {$ifdef ISACAN }   CANREAD := isa_regread(PortBase, reg_adr) ; {$endif}
        {$ifdef DONGLEMUX} CANREAD := mux_regread(PortBase, reg_adr) ; {$endif}
        {$ifdef DONGLEEPP} CANREAD := epp_regread(PortBase, reg_adr) ; {$endif}
      END { "CANREAD" } ;

    {*** CANWRITE(): ein Register des 82C200 schreiben ***}
    PROCEDURE CANWRITE(reg_adr : byte ; reg_val : byte) ;
      BEGIN
        {* Auswahl der anzusteuernden CAN-Hardware: ISA-Einsteckkarte oder
         * eine der Donglevarianten *}
        {$ifdef ISACAN }   isa_regwrite(PortBase, reg_adr, reg_val) ;  {$endif}
        {$ifdef DONGLEMUX}  mux_regwrite(PortBase, reg_adr, reg_val) ; {$endif}
        {$ifdef DONGLEEPP}  epp_regwrite(PortBase, reg_adr, reg_val) ; {$endif}
      END { "CANWRITE" } ;

    {*** CANBITSET(): die Bits in 'mask' im Controllerregister 'reg_adr' setzen ***}
    PROCEDURE CANBITSET(reg_adr: byte ; mask: byte) ;
      BEGIN
        {* Besonderheit des COMMAND-Registers: Bits 4..0 sind write only *}
        IF ( reg_adr = _COMMAND )
          THEN CANWRITE(_COMMAND, (CANREAD(_COMMAND) AND $E0) OR mask)
          ELSE CANWRITE(reg_adr, CANREAD(reg_adr) OR mask) ;
      END { "CANBITSET" } ;

    {*** CANBITCLR(): die Bits in 'mask' im Controllerregister 'reg_adr' loeschen ***}
    PROCEDURE CANBITCLR(reg_adr: byte ; mask: byte) ;
      BEGIN
        {* Besonderheit des COMMAND-Registers: Bits 4..0 sind write only *}
        IF ( reg_adr = _COMMAND )
          THEN CANWRITE(_COMMAND, (CANREAD(_COMMAND) AND $E0) AND NOT mask)
          ELSE CANWRITE(reg_adr, CANREAD(reg_adr) AND NOT mask) ;
      END { "CANBITCLR" } ;



    {************************************************************************}
    {***    Global verwendbare Prozeduren (Methoden) dieses Moduls        ***}
    {************************************************************************}


    {************************************************************************
     *      CAN_Read()
     *      Polling-Routine:
     *      -  empfangene Message in den Puffer 'PCANMsg->' uebertragen
     *      -  Ueberwachung des Bus-Status und ggf. Reset des CAN-Controllers.
     *
     *      Algorithmus angelehnt an [AMC], S. 18.
     *
     *      Parameter:
     *               PCANMsg: Zeiger auf Messagepuffer fuer die Rueckgabe der
     *               empfangenen Message.
     *
     *      Rueckgabe: Bitmaske mit Fehlern.
     *              0 :     Kein Fehler, neue Message steht in 'PCANMsg->'.
     *              CAN_ERR_RCVEMPTY: es wurde keine neue Message empfangen.
     *              CAN_ERR_OVERRUN : eine empfangene Message wurde nicht rechtzeitig
     *                              ausgelesen.
     *              CAN_ERR_BUSERROR: Controller meldet viele Busstoerungen.
     *              CAN_ERR_BUSOFF: Controller war Bus-Off und wurde wieder
     *                      aktiviert.
     *      *}
    FUNCTION CAN_Read(VAR CANMsg: TCANMsg) : integer ;
      LABEL 999 ;
      VAR
        received: boolean ;
        returnval: integer ;
      BEGIN
        {************************ Datenempfang **************************}
        returnval := 0 ;
        received := false ;
        IF ( CANREAD(_STATUS) AND $01 ) <> 0 {* Receive buffer status == full ? *}
          THEN BEGIN
              {***  empfange jede Message, die dem Akzeptanzfilter genuegt *}
              received := true ;
              CANMsg.ID := ( word(CANREAD(_RCVID)) shl 3) OR (CANREAD(_RCVLEN) shr 5) ;
              CANMsg.LEN := CANREAD(_RCVLEN) AND $0f ;
              IF (CANREAD(_RCVLEN) AND $10 ) <> 0
                THEN CANMsg.RTR := true  {* RTR-Bit gesetzt! *}
                ELSE BEGIN
                    {* den ganzen Empfangspuffer nach CAN_RWBUFF bewegen *}
                    CANMsg.RTR := false ;
                    CANMsg.DATA[0] := CANREAD(_RCVBUF) ;
                    CANMsg.DATA[1] := CANREAD(_RCVBUF+1) ;
                    CANMsg.DATA[2] := CANREAD(_RCVBUF+2) ;
                    CANMsg.DATA[3] := CANREAD(_RCVBUF+3) ;
                    CANMsg.DATA[4] := CANREAD(_RCVBUF+4) ;
                    CANMsg.DATA[5] := CANREAD(_RCVBUF+5) ;
                    CANMsg.DATA[6] := CANREAD(_RCVBUF+6) ;
                    CANMsg.DATA[7] := CANREAD(_RCVBUF+7) ;
                  END { else } ;
              {* release receive buffer *}
              CANBITSET(_COMMAND, $04) ;
              {* Evtl. 'Overrun'-Bedingung loeschen *}
              IF ( CANREAD(_STATUS) AND $02 ) <> 0THEN BEGIN
                    returnval := returnval OR CAN_ERR_OVERRUN ;
                    CANBITSET(_COMMAND, $08 ) ;
                  END ;
            END { then } ;

        {********************* Fehlerverarbeitung ***********************}
        IF ( CANREAD(_STATUS) AND $80 ) <> 0
          THEN BEGIN
              {*** Bus-Off: Controller is NOT involved in bus activities *}
              returnval := returnval OR CAN_ERR_BUSOFF ;
              {*** Reset request auf 'absent' *}
              CANBITCLR(_CONTROL, $01) ;
              {* Der Controller geht wieder 'On-Bus' nach dem Empfang
               * von 128 * 11 aufeinanderfolgenden rezessiven Bits.
               *}
            END { then } ELSE BEGIN
              {*** Bus-On: Controller is involved in bus activities **}

              IF ( CANREAD(_STATUS) AND $40 ) <> 0
                    {*** Ein Errorcounter hat sein Limit erreicht *}
                THEN returnval := returnval OR CAN_ERR_BUSERROR ;
            END { else } ;

        IF ( NOT received )
          THEN returnval := returnval OR CAN_ERR_RCVEMPTY ;

        CAN_Read := returnval ;
      END { "CAN_Read" } ;


    {************************************************************************
     *      CAN_Write()
     *      Message mit einer bestimmten Kennung auf den CAN-Bus schicken.
     *      Die Message aus 'PCANMsg->' uebernehmen.
     *      Algorithmus angelehnt an [AMC], S. 12
     *
     *      Parameter:
     *               PCANMsg: Zeiger auf Messagepuffer fuer die Rueckgabe der
     *               empfangenen Message.
     *      Rueckgabe: Bitmaske mit Fehlern
     *              0: Message wurde erfolgreich abgesetzt.
     *              CAN_ERR_XMTFULL: Es ist noch eine Uebertragung aktiv,
     *                      Message wurde NICHT abgesetzt.
     *      *}
    FUNCTION  CAN_Write(VAR CANMsg: TCANMsg) : integer ;
      LABEL 999 ;
      VAR
        tmp: byte ;
        returnval : integer ;
      BEGIN
        returnval := 0 ;
        {* Check 'Transmit Buffer Access'  *}
        IF (CANREAD(_STATUS) AND $04) = 0
          THEN {* transmit buffer locked *}
            BEGIN returnval := CAN_ERR_XMTFULL ; GOTO 999 END ;
        IF ( CANREAD(_STATUS) AND $20 ) <> 0
          THEN BEGIN returnval := CAN_ERR_XMTFULL ; GOTO 999 END ;

        {* Transmit Puffer fuellen *}
        CANWRITE(_XMTID, CANMsg.ID shr 3 ) ;
        {* montiere das Laengenbyte der Message *}
        tmp := ((CANMsg.ID AND $07) shl 5) OR (CANMsg.LEN AND $0f) ;
        IF ( CANMsg.RTR )
              {* sende remote-request-Message *}
          THEN CANWRITE( _XMTLEN, tmp OR $10 )
          ELSE BEGIN
              {* sende Datenmessage *}
              CANWRITE( _XMTLEN, tmp ) ;

              {* den ganzen Sendepuffer in den Controller bewegen *}
              CANWRITE(_XMTBUF+0, CANMsg.DATA[0]) ;
              CANWRITE(_XMTBUF+1, CANMsg.DATA[1]) ;
              CANWRITE(_XMTBUF+2, CANMsg.DATA[2]) ;
              CANWRITE(_XMTBUF+3, CANMsg.DATA[3]) ;
              CANWRITE(_XMTBUF+4, CANMsg.DATA[4]) ;
              CANWRITE(_XMTBUF+5, CANMsg.DATA[5]) ;
              CANWRITE(_XMTBUF+6, CANMsg.DATA[6]) ;
              CANWRITE(_XMTBUF+7, CANMsg.DATA[7]) ;
            END { else } ;
        {* transmit data by setting of TRANSMISSION REQUEST *}
        CANBITSET(_COMMAND, $01) ;
999:    ;
        CAN_write := returnval ;
      END { "CAN_Write" } ;


    {*******************************************************************
     *      CAN_Status()
     *      Abfrage des Statusregisters des 82C200 und Rueckgabe
     *      einer oder mehrere CAN_ERR_xxx Bedingungen.
     *      *}
    FUNCTION  CAN_Status : integer ;
      VAR
        tmp: byte ;
        returnval: integer ;
      BEGIN
        returnval := 0 ;
        tmp := CANREAD(_STATUS) ;

        IF  (tmp AND $01) = 0
          THEN returnval := returnval OR CAN_ERR_RCVEMPTY ;
        IF ( tmp AND $02 ) <> 0
          THEN returnval := returnval OR CAN_ERR_OVERRUN ;
        IF ( tmp AND $04 ) = 0
          THEN returnval := returnval OR CAN_ERR_XMTFULL ; {* transmit buffer locked *}
        IF ( tmp AND $20 ) <> 0
          THEN returnval := returnval OR CAN_ERR_XMTFULL ;
        IF ( tmp AND $40 ) <> 0
          THEN returnval := returnval OR CAN_ERR_BUSERROR;
        IF ( tmp AND $80 ) <> 0
          THEN returnval := returnval OR CAN_ERR_BUSOFF ;

        CAN_Status := returnval ;
      END { "CAN_Status" } ;


    {*******************************************************************
     *      CAN_Init()
     *      Wird nach dem Anschalten einmal durchlaufen,
     *      und bei Re-Initialisierung.
     *      Fuehrt Registertest auf dem 82C200 aus,
     *      programmiert dann Baudrate, Konfiguration der Sende/Empfangstreiber.
     *      Die Acceptance-Mask wird fuer den Empfang aller Messages gesetzt.
     *      Parameter:
     *              PortBase: Basisadresse der installierten Hadrware.
     *              btr0btr1: Wert fuer die Timingregister des 82C200.
     *      *}
    FUNCTION CAN_Init(_portbase : word; btr0btr1 : word) : integer ;
      LABEL 999 ;
      VAR
        i : integer ;
      BEGIN
        PortBase := _portbase ;
        CANHWINIT(PortBase) ;

        {* RESET, no Interrupts, Pin REF is Output,
         * Synch only by transitions from recessive to dominant  *}
        CANWRITE(_CONTROL, $01) ;

        {* Register 4..8 sind 8 Bit read/write; 20..29 ist XMT-Buffer *}
        FOR i := 4 TO 29 DO
          IF ( i < 9 ) OR ( i > 19 ) THEN BEGIN
                CANWRITE(i, i+5) ; { jedes Register mit 'Adresse+5' testen }
                IF ( CANREAD(i) <> i+5 )
                  THEN BEGIN CAN_Init := CAN_ERR_REGTEST ; GOTO 999 END ;
              END ;
        FOR i := 4 TO 29 DO
          IF ( i < 9 ) OR ( i > 19 ) THEN
              IF ( CANREAD(i) <> i+5 )
                THEN BEGIN CAN_Init := CAN_ERR_REGTEST ; GOTO 999 END ;

        CANWRITE(_COMMAND, $0E) ; {* CLr Overrun, Release RcvBuff, Abort Xmission *}

        {* Acceptance Code = Kennung fuer Empfang *}
        CANWRITE(_ACC_CODE, 0) ;
        CANWRITE(_ACC_MASK, $ff) ;   {* Kein Bit der Kennung ist relevant *}

        nomsgregistered := true ;        {* Flag fuer  'CAN_registerMsg()' *}

        {* TIMING0/1 fuer versch. ISO-Baudraten bei 16Mhz Quartz
         * [PLC] S.4, ausfuehrlich [TDS] S. 46f
         *}
        CANWRITE(_TIMING0, (btr0btr1 shr 8) AND $FF) ;
        CANWRITE(_TIMING1, btr0btr1 AND $FF) ;

        {* Output Control: Push/Pull an TX0, TX1 disabled
         * OCTP1 = OCTN1 = OCPOL1 = 0,
         * OCTP0 = OCTN0 = 1, OCPOL0 = 0 (keine Invertierung)
         * OCMODE1 = 1, OCMODE0 = 0  (normal)
         * [TDS] S.50 u. 119 *}
        CANWRITE(_OUTPUTCTRL, 0 OR $18 OR $02) ;

        CANWRITE(_CONTROL, $00) ; {* Clear reset-request *}

        controlleractive := true ;

        CAN_Init := CAN_ERR_OK ;
999:    ;
      END { "CAN_Init" } ;


    {*******************************************************************
     *      CAN_registerMsg()
     *      programmiert das Empfangsfilter so um, da zustzlich zu allen
     *      anderen Messages jetzt auch noch 'ID' empfangen wird.
     *
     *      - Beim ersten Aufruf von 'CAN_registerMsg()' wird vom Empfang
     *        aller Messages auf den Empfang eine einzigen umgeschaltet.
     *      - reset durch 'CAN_Init()'
     *      *}
    PROCEDURE CAN_registerMsg( ID: word ) ;
      VAR
        n : byte ;
      BEGIN
        {* RESET, no Interrupts, Pin REF is Output,
         * Synch only by transitions from recessive to dominant  *}
        CANWRITE(_CONTROL, $01) ;

        {* Algorithmus:
         * Acceptance-Code bleibt immer die beim 1. Aufruf angegebene ID.
         * Folgende Aufrufe loeschen nur noch Bits in der Acceptance-Mask.
         *}
        n := (ID shr 3) ;     {* nur die obersten 8 Bit der ID sind relevant *}
        IF nomsgregistered THEN BEGIN
              nomsgregistered := false ;
              CANWRITE(_ACC_CODE, n) ;
              CANWRITE(_ACC_MASK, 0) ; {* Alle Bits der ID sind relevant *}
            END ELSE BEGIN
              {* Check, in welchen Bits 'n' vom Acceptance-Code differiert.
               * Diese Bits zustzlich in der Acceptance-Mask setzen
               * (XOR ist 1 bei Differenz) *}
              n := n xor CANREAD(_ACC_CODE) ; {* n : high = differierende Bits *}
              n := n OR CANREAD(_ACC_MASK) ; {* n : high = neue 'dont care' Bits *}
              CANWRITE(_ACC_MASK, n) ;
            END { else } ;

        CANWRITE(_CONTROL, $00) ;      {* Reset Ende *}
      END { "CAN_registerMsg" } ;


    {*******************************************************************
     *      CAN_Close()
     *      Controller deaktivieren.
     *      Neustart dann mit CAN_Init() ;
     *      *}
    PROCEDURE CAN_Close ;
      BEGIN
        IF controlleractive
          THEN CANWRITE(_CONTROL, $01) ; {* Controller in Reset-Mode *}
        controlleractive := false ;
      END ;

    BEGIN
      controlleractive := false ;
    END .
