
UNIT DonglReg ;

  {
     DONGLREG.PAS
     Borland-Pascal / Delphi Unit zur Ansteuerung der
     Register des CAN-Controllers 82C200 auf dem
     PEAK-CAN-Dongle.

     Allgemeines zur Ansteuerung:
     ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
     Die Ansteuerung des PEAK-CAN-Dongles erfolgt ber den LPT-Port.
     Die Signaladern des LPT-Portes sind ber 3 Register im IO-Bereich
     des PC's ansprechbar. Die Register liegen bei folgenden Adressen:
     "A" = Basisadresse      (Basisadr. z.B. $378 bei LPT1)
     "B" = Basisadresse+1
     "C" = Basisadresse+2

     Funktion der Registerbits:

     Signal          Richtung        Port
     --------        ---------       -------------------
     DATA0..DATA7    O (Basis-Mode)  "A"
                     I/O (Bidi-Mode)
     STROBE*         O               "C", Bit 0
     ALF*            O               "C", Bit 1
     ERR*            I               "B", Bit 3
     INI*            O               "C", Bit 2
     DSL*            O               "C", Bit 3
     ACK*            I               "B", Bit 6 (lst Interrupt aus)
     BUSY            I               "B", Bit 7
     PEND            I               "B", Bit 5
     SEL             I               "B", Bit 4

     INT_ENABLE      (intern)        "C", Bit 4
     READ_ENABLE     (intern)        "C", Bit 5 (nur BIDI-Schnittstellen)

     Manche Port-C-Leitungen werden zustzlich invertiert,
     daher mssen Ausgaben nach Port C immer mit einem "XOR $0B"
     verrechnet werden.


     Konventionelle und bidirektionale LPT-Ports:
     ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
     Im allgemeinen ist DATA0..7 des LPT-Port nicht bidirektional.
     Ausgaben an den CAN-Controllers gehen ber DATA0..7,
     Ergebnisse werden in 2 4-Bit-Nibbles ber Port "B" rckgelesen.
     (sogenannter "Multiplex-Modus").

     Wurde im PC-BIOS der LPT-Port auf "EPP" oder "EPC" gestellt,
     kann der LPT-Port ber das ECR-Register bei Basisadresse+$402
     in den PS/2-Modus geschaltet werden.
     Dann knnen Daten schneller ber DATA0..7 rckgelesen werden.


     Allgemeines zum CAN-Controller:
     ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
     Der CAN-Controller 82C200 besitzt 32 interne Register.
     Diese sind 0..31 nummeriert.
     Ein Schreiben nach Register 31 lst einen Hardwarereset
     des 82C200 aus.


     Copyright     (c) 1999 PEAK-System Technik GmbH, Darmstadt.
  }

  INTERFACE

    { bergabeparameter:
      'basisadr': Adresse des LPT-Ports (z.B. '$378' fr LPT1).
      'reg_adr':  anzusprechendes Register im 82C200
                  (z.B. '2' fr das Statusregister).
      'reg_val':  in ein Register zu schreibender Wert.
    }

    { Schreib/Leseroutinen fuer den MUX-Modus }
    PROCEDURE mux_reginit(basisadr: word) ;
    FUNCTION  mux_regread(basisadr: word ; reg_adr : byte) : byte ;
    PROCEDURE mux_regwrite(basisadr: word; reg_adr : byte ; reg_val : byte) ;

    { Schreib/Leseroutinen fr den EPP-Modus }
    PROCEDURE epp_reginit(basisadr: word) ;
    FUNCTION  epp_regread(basisadr: word ; reg_adr : byte) : byte ;
    PROCEDURE epp_regwrite(basisadr: word; reg_adr : byte ; reg_val : byte) ;


  IMPLEMENTATION


    CONST
      { -- Tabelle, die den an PB3..PB7 gelesenen Bits
           das originale Nibble am Ausgang des Multiplexers zuordnet.
           Besonderheit: PB6 wird ignoriert. }
      nibble_decode : ARRAY[0..31] OF byte = (
            $08, $09, $0a, $0b, $0c, $0d, $0e, $0f,
            $08, $09, $0a, $0b, $0c, $0d, $0e, $0f,
            $00, $01, $02, $03, $04, $05, $06, $07,
            $00, $01, $02, $03, $04, $05, $06, $07 ) ;


    PROCEDURE mux_reginit(basisadr: word) ;
       BEGIN
       END ;


    FUNCTION mux_regread(basisadr: word ; reg_adr : byte) : byte ;
      VAR
        PA, PB, PC : word ;
        b0, b1 : byte ;
      BEGIN
        PA := basisadr ; PB := basisadr+1 ; PC := basisadr+2 ;

        { -- Strobe (PC0) H->L: Adressuebernahme.
             Strobe L->H: Busfreigabe
             Mode (PC1) = L: MUX-Mode
             Nibbleumschaltung in der Lesephase mit PA6:
             PA6=0: PB3,PB4,PB5,PB7 = DATA4..7
             PA6=1:   "   "  "   "  = DATA0..3
        }
        { Bits in PC sind:  0 0 0 0 ispEN x Mode Strb Bemerkung}
        {                           ~~~~  ~ ~~~~~~~~~~~~~~~~~~~~~~~~}
        Port[PC] := $0B xor $1D ; { H     H L    H    Grundstellung }
        Port[PA] := (reg_adr AND $1F) OR $80 ; { Bit 7 =H : Lesezyklus  }
        Port[PC] := $0B xor $1C ; { H     H L    L    Adr-Uebernahme }

        b1 := nibble_decode[Port[PB] shr 3] ; {     PA6=0: High Nibble lesen }
        Port[PA] := $40 ;
        b0 := nibble_decode[Port[PB] shr 3] ; {     PA6=1: Low Nibble lesen }
        Port[PC] := $0B xor $1D ; { H     H L    H    Grundstellung }
        mux_regread := (b1 shl 4) OR b0 ;
      END { "mux_regread" } ;



    PROCEDURE mux_regwrite(basisadr: word; reg_adr : byte ; reg_val : byte) ;
      VAR PA, PC : word ;
      BEGIN
        PA := basisadr; PC := basisadr+2 ;
        { -- Strobe (PC0) H->L: Adressuebernahme.
             Strobe L->H: Busfreigabe
             Mode (PC1) = L: MUX-Mode }
        { Bits in PC sind:  0 0 0 0 ispEN x Mode Strb Bemerkung}
        {                           ~~~~  ~ ~~~~~~~~~~~~~~~~~~~~~~~~}
        Port[PC] := $0B xor $1D ; { H     H L    H    Grundstellung }
        Port[PA] := (reg_adr AND $1F);{               Bit 7 =0 : Schreibzyklus  }
        Port[PC] := $0B xor $1C ; { H     H L    L    Adr-Uebernahme }
        Port[PA] := reg_val ;     { Schreibdaten ausgeben und ...   }
        Port[PC] := $0B xor $1D ; { H     H L    H    ... uebernehmen}
        {                                        Und wieder in Grundstellung }
      END { "mux_regwrite" } ;



    PROCEDURE epp_reginit(basisadr: word) ;
       BEGIN
        { 1284-kompatiblen Port in den "PS/2"-Mode setzen.
          Damit das ECR-Register bei Offset $402 freigeschaltet ist,
          mu im BIOS "EPP/EPC" eingestellt werden sein. }
        Port[basisadr+$402] := (Port[basisadr+$402] and $1F) or $20 ;
       END ;

    FUNCTION epp_regread(basisadr: word ; reg_adr : byte) : byte ;
      VAR
        PA, PB, PC : word ;
      BEGIN
        PA := basisadr ; PB := basisadr+1 ; PC := basisadr+2 ;

        { -- Strobe (PC0) H->L: Adressuebernahme.
             Strobe L->H: Busfreigabe
             Mode (PC1) = H EPP-Mode }
        { Bits in PC sind:  0 0 0 0 ispEN x Mode Strb Bemerkung}
        {                           ~~~~  ~ ~~~~~~~~~~~~~~~~~~~~~~~~}
        Port[PC] := $0B xor $1F ; { H     H H    H    Grundstellung }
        Port[PA] := (reg_adr AND $1F) OR $80 ; { Bit 7 =H : Lesezyklus  }
        Port[PC] := $0B xor $3E ; { H     H H    L    Adr-Uebernahme, Tristate }
        epp_regread := Port[PA] ;
        Port[PC] := $0B xor $1F ; { H     H H    H    Grundstellung }
      END { "epp_regread" } ;


    PROCEDURE epp_regwrite(basisadr: word; reg_adr : byte ; reg_val : byte) ;
      VAR
        PA, PC : word ;
      BEGIN
        PA := basisadr ; PC := basisadr+2 ;
        { -- Strobe (PC0) H->L: Adressuebernahme.
             Strobe L->H: Busfreigabe
             Mode (PC1) = H EPP-Mode }
        { Bits in PC sind:  0 0 0 0 ispEN x Mode Strb Bemerkung}
        {                           ~~~~  ~ ~~~~~~~~~~~~~~~~~~~~~~~~}
        Port[PC] := $0B xor $1F ; { H     H H    H    Grundstellung }
        Port[PA] := (reg_adr AND $1F);{ Bit 7 =0 : Schreibzyklus  }
        Port[PC] := $0B xor $1E ; { H     H H    L    Adr-uebernahme }
        Port[PA] := reg_val;      { Schreibdaten ausgeben und ...   }
        Port[PC] := $0B xor $1F ; { H     H H    H    ... uebernehmen}
        {                                             Und wieder in Grundstellung }
      END { "epp_regwrite" } ;

  END { "DonglReg" } .
